unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DSound, afDXSPlay, afDXSRec, afDXSUtil, StdCtrls, Buttons;

type
  TfrmMain = class(TForm)
    MemoStatus: TMemo;
    btnSpeaker: TSpeedButton;
    btnMicrophone: TSpeedButton;
    Label1: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnMicrophoneClick(Sender: TObject);
    procedure btnSpeakerClick(Sender: TObject);
  private
    { Private declarations }
    Buff : TafDXSPlayStreamingBuffer;
    RecBuff : TafDXSRecStreamingBuffer;
    procedure DataAvailable(Sender : TObject);
    procedure DataRequired(Sender : TObject);
  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.DFM}

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  // Initialize then DX Playback object
  DXSPlay_Init(Self, Handle, DSSCL_PRIORITY);
  if Assigned(DXSPlay) then DXSPlay.SetPrimaryFormat(1, 8, 22050);
  // set the primary format ( a good thing to do )
  if (Assigned(DXSPlay)) and (DXSPlay.Ready) then MemoStatus.Lines.Add('DX Sound Playback available.')
    else MemoStatus.Lines.Add('DX Sound Playback not available.');

  // Initialize the DX Capture object
  DXSRec_Init(Self);
  if Assigned(DXSRec) and (DXSRec.Ready) then begin
    MemoStatus.Lines.Add('DX Sound Capture available.');
  end else begin
    MemoStatus.Lines.Add('DX Sound Capture not available.');
    MemoStatus.Lines.Add('You need a full-duplex sound card.');
    btnMicrophone.Enabled := False;
  end;
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  DXSPlay_Free; // Free the Playback object
  DXSRec_Free;  // Free the Capture cbject
end;

procedure TfrmMain.btnMicrophoneClick(Sender: TObject);
begin
  if btnMicrophone.Down then begin
    if Assigned(DXSRec) and DXSRec.Ready then begin
      // aquire the streaming capture buffer
      RecBuff := DXSRec.NewStreamingBuffer(1, 8, 22050, 0, 20000);
      if Assigned(RecBuff) then begin
        // tell it what to call when data is ready for reading
        RecBuff.OnDataAvailable := DataAvailable;
        // start the buffer capturing
        RecBuff.RecLoop;
        MemoStatus.Lines.Add('DX Sound Capture Buffer recording.');
      end else begin
        MemoStatus.Lines.Add('DX Sound Capture Buffer recording FAILED.');
      end;
    end;
  end else begin
    if Assigned(DXSRec) then
      if Assigned(RecBuff) then begin
        // stop the capture and
        RecBuff.Stop;
        // release the buffer
        RecBuff.Free;
        MemoStatus.Lines.Add('DX Sound Capture stopped.');
      end;
  end;
end;

procedure TfrmMain.btnSpeakerClick(Sender: TObject);
begin
  if btnSpeaker.Down then begin
    if Assigned(DXSPlay) then begin
      // create the playback buffer
      Buff := DXSPlay.NewStreamingBuffer(1, 8, 22050, 0, 20000);
      if Assigned(Buff) then begin
        // tell the buffer who to call for data required
        Buff.OnDataRequired := DataRequired;
        // start the buffer playing
        Buff.PlayLoop;
        MemoStatus.Lines.Add('DX Sound Playback Buffer Playing.');
      end;
    end;
  end else begin
    if Assigned(Buff) then begin
      // stop and free the playback buffer
      Buff.Stop;
      Buff.Free;
      Buff := nil;
    end;
    MemoStatus.Lines.Add('DX Sound Playback stopped.');
  end;
end;


procedure TfrmMain.DataAvailable(Sender : TObject);
var
  Ptr : Pointer;
  Size : Longint;
begin
  MemoStatus.Lines.Add('Data Available');

  // get data from mic and pass it to the speaker
  Ptr := nil;
  Size := 0;
  RecBuff.RetrieveData(Ptr, Size);  // get the data available

  if Assigned(Buff) then            // if there's an out buffer
    Buff.InsertData(Ptr, Size);     // feed it to the out buffer

  FreeMem(Ptr, Size);                // free the used memory
end;

procedure TfrmMain.DataRequired(Sender : TObject);
begin
  MemoStatus.Lines.Add('Need Data');
end;

end.
